from experiment_automator.exceptions import ImageUploaderException
from experiment_automator.constants import SlackConstants, OAuthConstants, FlickrConstants, OtherConstants
from experiment_automator.oauth_client import OAuthClient
from experiment_automator.utils import DebugLogCat
from xml.etree import ElementTree


class ImageUploader:
    def __init__(self, debug, image_uploader_config):
        self.debug = debug

        if image_uploader_config is None:
            raise ImageUploaderException("%s - Image uploader config is None. Please specify api key in config!" % self.__class_name())

        self.image_uploader_config = image_uploader_config
        self.oauth_config = image_uploader_config.get(OAuthConstants.KEY_OAUTH)

        # If Flickr service defined get configurations from Constants
        if not (self.oauth_config is None) and self.oauth_config.get(OAuthConstants.KEY_OAUTH_SERVICE_NAME) == OtherConstants.IMAGE_SERVICE_FLICKR:
            self.oauth_config[OAuthConstants.KEY_OAUTH_REQUEST_TOKEN_URL] = FlickrConstants.FLICKR_OAUTH_REQUEST_TOKEN_URL
            self.oauth_config[OAuthConstants.KEY_OAUTH_AUTHORIZATION_URL] = FlickrConstants.FLICKR_OAUTH_AUTHORIZATION_URL
            self.oauth_config[OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_URL] = FlickrConstants.FLICKR_OAUTH_ACCESS_TOKEN_URL
            self.oauth_config[OAuthConstants.KEY_OAUTH_BASE_URL] = FlickrConstants.FLICKR_OAUTH_BASE_URL

            self.image_uploader_config[SlackConstants.KEY_SLACK_IMAGE_SERVICE_UPLOAD_URL] = FlickrConstants.FLICKR_UPLOAD_URL

    def __class_name(self):
        return self.__class__.__name__

    def get_session(self, callback, **params):
        return OAuthClient(self.debug, self.image_uploader_config.get(OAuthConstants.KEY_OAUTH)).connect(callback, **params)


class FlickrImageUploader(ImageUploader):
    def __init__(self, debug, image_uploader_config):
        super().__init__(debug, image_uploader_config)
        self.session = super().get_session(OAuthConstants.KEY_OAUTH_CALLBACK_OOB, perms="write")

    def __class_name(self):
        return self.__class__.__name__

    def __get_image_static_url(self, photo_id, image_type):
        data = {
            "photo_id": photo_id,
            "method": FlickrConstants.FLICKR_GET_IMAGE_SIZES_METHOD
        }

        response = self.session.post(self.oauth_config.get(OAuthConstants.KEY_OAUTH_BASE_URL), data=data).text

        sizes = {}
        for size in ElementTree.fromstring(response).find('sizes').findall('size'):
            sizes[size.attrib["label"]] = size.attrib["source"]

        return sizes[image_type]

    def upload_image(self, image_path, image_type):
        DebugLogCat.log(self.debug, self.__class_name(), "Uploading image \"%s\"" % image_path)

        files = {
            "photo": open(image_path, 'rb'),
        }

        photo_id = ElementTree.fromstring(self.session.post(self.image_uploader_config.get(SlackConstants.KEY_SLACK_IMAGE_SERVICE_UPLOAD_URL), files=files).text).find("photoid").text

        if photo_id is None:
            DebugLogCat.log(self.debug, self.__class_name(), "Image upload could not be completed!" % photo_id)
        else:
            DebugLogCat.log(self.debug, self.__class_name(), "Image upload completed. Photo id: \"%s\" !" % photo_id)

        return self.__get_image_static_url(photo_id, image_type)



# TODO Write warnings for debugging and info user
# TODO Parallel model training
# TODO Database integration (Document store)
# TODO Distributed task queue (Celery) integration for async tasks

# ---------------------- Next features ----------------------
# TODO Different notification channels
# TODO HTTP Communication
# TODO Web based dashboard (Plotly visualization)


from itertools import product
from numpy import arange
from datetime import datetime
from typing import Callable
from time import time, gmtime, strftime
import traceback

# ---------------------------------------------------------------
from experiment_automator.constants import ExperimentConstants, SlackConstants, CSVReporterConstants, OtherConstants
from experiment_automator.slack_notifier import SlackNotifier
from experiment_automator.csv_reporter import CSVReporter
from experiment_automator.utils import DebugLogCat
from experiment_automator.config import Config
from experiment_automator.result_container import ResultContainer
from experiment_automator.exceptions import *


class ExperimentAutomator:
    def __init__(self, config_filename):
        self.config_filename = config_filename
        self.experiment_params = []

        try:
            self.config = Config(config_filename)
            self.debug = self.config.get(OtherConstants.KEY_DEBUG)
            self.debug_traceback = self.config.get(OtherConstants.KEY_DEBUG_TRACEBACK)
            self.slack_notifier = SlackNotifier(self.debug,
                                                self.config.get_dict(SlackConstants.KEY_SLACK))
            self.csv_logger = CSVReporter(self.debug,
                                          self.config.get_dict(ExperimentConstants.KEY_EXPERIMENT),
                                          self.config.get_dict(CSVReporterConstants.KEY_CSV))
            self.__prepare_experiment_params()
            self.__gen_experiment_param_combinations()
        except ConfigNotFoundException as ex:
            print(ex)
            self.__print_stacktrace()
        except ImageUploaderException as ex:
            print(ex)
            self.__print_stacktrace()

    def __print_stacktrace(self):
        if self.debug_traceback:
            traceback.print_exc()

    def __class_name(self):
        return self.__class__.__name__
    
    def __prepare_experiment_params(self):
        DebugLogCat.log(self.debug, self.__class_name(), "Preparing experiment parameters for training...!")

        experiment_config = self.config.get_dict(ExperimentConstants.KEY_EXPERIMENT)

        if experiment_config is None:
            raise ConfigNotFoundException("%s - Experiment configuration not found in config \"%s\"" % (self.__class_name(), self.config_filename))

        experiment_params = experiment_config.get(ExperimentConstants.KEY_EXPERIMENT_PARAMS, None)

        if experiment_params is None:
            raise ConfigNotFoundException("%s - Experiment parameters configurations not found in config \"%s\"" % (self.__class_name(), self.config_filename))

        for (experiment_param_name, experiment_param_val) in experiment_params.items():
            # If parameter types includes range specification, generate from that range specification
            if isinstance(experiment_param_val, dict) and experiment_param_val.get(ExperimentConstants.KEY_EXPERIMENT_PARAM_RANGE, None) is not None:
                range_spec = experiment_param_val.get(ExperimentConstants.KEY_EXPERIMENT_PARAM_RANGE)

                # Generating values from range specification
                gen_values = []
                for val in arange(range_spec[0], range_spec[1] + range_spec[2], range_spec[2]):
                    gen_values.append(val)

                experiment_params[experiment_param_name] = gen_values

    def __gen_experiment_param_combinations(self):
        DebugLogCat.log(self.debug, self.__class_name(), "Generating experiment parameters combinations for training...!")

        experiment_params = self.config.get_dict(ExperimentConstants.KEY_EXPERIMENT).get(ExperimentConstants.KEY_EXPERIMENT_PARAMS)
        experiment_param_vals = []

        for (_, experiment_param_val) in experiment_params.items():
            experiment_param_vals.append(experiment_param_val)

        # Generating combinations of experiment attributes
        for experiment_attr_combination in product(*experiment_param_vals):
            self.experiment_params.append(dict(zip(experiment_params.keys(), experiment_attr_combination)))

    # Driver of the program
    def run(self, fn: Callable[[dict, ResultContainer], None]):
        for attrs in self.experiment_params:
            try:
                start = time()
                start_time = datetime.now().strftime("%m/%d/%y %H:%M:%S")

                results = ResultContainer(attrs)

                # Execute ML pipeline given by user
                fn(attrs, results)

                end = time()

                completion_time = datetime.now().strftime("%m/%d/%y %H:%M:%S")
                elapsed_time = strftime("%H:%M:%S", gmtime(end - start))

                common_data = {
                    OtherConstants.KEY_START_TIME: start_time,
                    OtherConstants.KEY_COMPLETION_TIME: completion_time,
                    OtherConstants.KEY_ELAPSED_TIME: elapsed_time,
                    OtherConstants.KEY_ERROR_CAUSE: "-"
                }

                experiment_results_payload = results.get_experiment_results_payload()
                slack_payload = results.get_slack_payload()

                # Merge parameters with Slack payload
                slack_payload.update({ExperimentConstants.KEY_EXPERIMENT_PARAMS: attrs})
                # Merge other infos with Slack payload
                slack_payload.get(ExperimentConstants.KEY_EXPERIMENT_RESULTS).update(common_data)

                # Merge parameters with experiment results payload
                experiment_results_payload.update(attrs)
                # Merge other infos with experiment results payload
                experiment_results_payload.update(common_data)
                # Additional exec status added to experiment results payload
                experiment_results_payload.update({OtherConstants.KEY_EXEC_STATUS: OtherConstants.EXECUTION_STATUS_SUCCESS})

                DebugLogCat.log(self.debug, self.__class_name(), "Model training is completed successfully. Sending notification to Slack channel!")

                DebugLogCat.log(self.debug, self.__class_name(), "Slack payload: %s" % slack_payload)
                DebugLogCat.log(self.debug, self.__class_name(), "Model results payload: %s" % experiment_results_payload)

                # If model training completes successfully send notification to Slack channel, save result to log file
                # and send them to Drive
                self.slack_notifier.notify(SlackConstants.KEY_SLACK_NOTIFICATION_SUCCESS, slack_payload)
                self.csv_logger.save_results_to_csv(experiment_results_payload)
            except ConfigNotFoundException as ex:
                print(ex)
                self.__print_stacktrace()
                break
            except WorkingDirectoryDoesNotExistException as ex:
                print(ex)
                self.__print_stacktrace()
                break
            except Exception as ex:
                print(ex)
                self.__print_stacktrace()

                end = time()

                common_data = {
                    OtherConstants.KEY_START_TIME: start_time,
                    OtherConstants.KEY_COMPLETION_TIME: datetime.now().strftime("%m/%d/%y %H:%M:%S"),
                    OtherConstants.KEY_ELAPSED_TIME: strftime("%H:%M:%S", gmtime(end - start)),
                    OtherConstants.KEY_ERROR_CAUSE: "\"%s\"" % str(ex).replace("\n", " ")
                }

                # Merge parameters and other infos with experiment results payload
                experiment_results_payload = {}
                experiment_results_payload.update(attrs)
                experiment_results_payload.update(common_data)
                experiment_results_payload.update({OtherConstants.KEY_EXEC_STATUS: OtherConstants.EXECUTION_STATUS_FAILED})

                # Merge parameters and other infos with Slack payload
                slack_payload = {
                    ExperimentConstants.KEY_EXPERIMENT_PARAMS: attrs,
                    ExperimentConstants.KEY_EXPERIMENT_RESULTS: common_data
                }

                DebugLogCat.log(self.debug, self.__class_name(), "Model training couldn't completed successfully. Sending notification to Slack channel!")

                # If model training and evaluation terminates with error status send notification
                self.slack_notifier.notify(SlackConstants.KEY_SLACK_NOTIFICATION_FAIL, slack_payload)
                self.csv_logger.save_results_to_csv(experiment_results_payload)
            finally:
                self.experiment_params.pop()



from experiment_automator.constants import SlackConstants, OtherConstants, FlickrConstants, ExperimentConstants
from experiment_automator.utils import DebugLogCat
from experiment_automator.image_uploader import FlickrImageUploader
from experiment_automator.exceptions import ImageUploaderException
from requests import post


class SlackNotifier:
    def __init__(self, debug, slack_config):
        self.debug = debug
        self.slack_config = slack_config

        if self.slack_config.get(SlackConstants.KEY_SLACK_IMAGE_SERVICE, None) is None:
            raise ImageUploaderException("%s - Image service configuration is None. Please specify image service informations in config!" % self.__class_name())

        self.image_uploader = FlickrImageUploader(self.debug, self.slack_config.get(SlackConstants.KEY_SLACK_IMAGE_SERVICE))

    def __class_name(self):
        return self.__class__.__name__

    def __create_slack_attachment(self, attachments, status, pretext=None, title=None, main_image=None, fields=None, footer=SlackConstants.VALUE_SLACK_NOTIFICATION_FOOTER):
        # Default attributes of attachments
        attachment = {
            SlackConstants.KEY_SLACK_NOTIFICATION_COLOR: SlackConstants.VALUE_SLACK_NOTIFICATION_COLORS[status],
            SlackConstants.KEY_SLACK_NOTIFICATION_FOOTER: footer
        }

        # Set pretext of attachment if you want
        if not (pretext is None):
            attachment[SlackConstants.KEY_SLACK_NOTIFICATION_PRETEXT] = SlackConstants.VALUE_SLACK_NOTIFICATION_PRETEXTS[status]

        # Set title of attachment if you want
        if not (title is None):
            attachment[SlackConstants.KEY_SLACK_NOTIFICATION_TITLE] = title

        # Set main image of notification if it is available
        if not (main_image is None):
            image_url = self.image_uploader.upload_image(image_path=main_image, image_type=FlickrConstants.FLICKR_IMAGE_TYPE_ORIGINAL)
            DebugLogCat.log(self.debug, self.__class_name(), "Retrieved url of figure: %s" % image_url)
            attachment[SlackConstants.KEY_SLACK_NOTIFICATION_IMAGE_URL] = image_url

        # Adding notification fields
        if not (fields is None):
            attachment[SlackConstants.KEY_SLACK_FIELDS] = []

            for (title, value) in fields.items():
                if (title in [OtherConstants.KEY_EXEC_STATUS, SlackConstants.KEY_SLACK_MAIN_IMAGE,SlackConstants.KEY_SLACK_IMAGE_ATTACHMENTS]) or \
                   (status == SlackConstants.KEY_SLACK_NOTIFICATION_SUCCESS and title == OtherConstants.KEY_ERROR_CAUSE):
                    continue

                attachment[SlackConstants.KEY_SLACK_FIELDS].append({
                    SlackConstants.KEY_SLACK_FIELD_TITLE: title,
                    SlackConstants.KEY_SLACK_FIELD_VALUE: str(value),
                    SlackConstants.KEY_SLACK_FIELD_SHORT: True
                })

        attachments.append(attachment)

    def __create_slack_image_attachments(self, attachments, status, images):
        if not (images is None) and status == SlackConstants.KEY_SLACK_NOTIFICATION_SUCCESS:
            DebugLogCat.log(self.debug, self.__class_name(), "There is a figures in Slack payload named \"%s\"!" % images)

            for (label, image_path) in images:
                image_url = self.image_uploader.upload_image(image_path=image_path, image_type=FlickrConstants.FLICKR_IMAGE_TYPE_ORIGINAL)

                DebugLogCat.log(self.debug, self.__class_name(), "Retrieved url of figure: %s" % image_url)

                attachments.append({
                    SlackConstants.KEY_SLACK_NOTIFICATION_TITLE: label,
                    SlackConstants.KEY_SLACK_NOTIFICATION_TITLE_LINK: image_url,
                    SlackConstants.KEY_SLACK_NOTIFICATION_THUMB_URL: image_url,
                    SlackConstants.KEY_SLACK_NOTIFICATION_COLOR: SlackConstants.VALUE_SLACK_NOTIFICATION_COLORS[status]
                })
        else:
            DebugLogCat.log(self.debug, self.__class_name(), "There is no figure in Slack payload!")

    def __generate_notification_data(self, status, payload):
        notification_data = {SlackConstants.KEY_SLACK_ATTACHMENTS: []}

        attachments = notification_data.get(SlackConstants.KEY_SLACK_ATTACHMENTS)

        # Create attachment for model results
        self.__create_slack_attachment(attachments=attachments,
                                       status=status,
                                       pretext=SlackConstants.VALUE_SLACK_NOTIFICATION_PRETEXTS[status],
                                       title=SlackConstants.VALUE_SLACK_NOTIFICATION_TITLE_RESULTS,
                                       main_image=payload.get(SlackConstants.KEY_SLACK_MAIN_IMAGE, None),
                                       fields=payload.get(ExperimentConstants.KEY_EXPERIMENT_RESULTS, None))

        # Create attachment for additional figures
        self.__create_slack_image_attachments(attachments=attachments,
                                              status=status,
                                              images=payload.get(SlackConstants.KEY_SLACK_IMAGE_ATTACHMENTS, None))

        # Create attachment for model parameters
        self.__create_slack_attachment(attachments=attachments,
                                       status=status,
                                       title=SlackConstants.VALUE_SLACK_NOTIFICATION_TITLE_PARAMS,
                                       fields=payload.get(ExperimentConstants.KEY_EXPERIMENT_PARAMS, None))

        DebugLogCat.log(self.debug, self.__class_name(), "Generated notification is \"%s\"" % str(notification_data))

        return notification_data

    def notify(self, status, payload):
        if not (self.slack_config is None) and not (self.slack_config.get(SlackConstants.KEY_SLACK_WEBHOOK_URL) is None):
            DebugLogCat.log(self.debug, self.__class_name(), "Slack configured properly. We send notification to Slack channel!")

            post(str(self.slack_config.get(SlackConstants.KEY_SLACK_WEBHOOK_URL)), json=self.__generate_notification_data(status, payload))
        else:
            DebugLogCat.log(self.debug, self.__class_name(), "There is no configuration for Slack. So we can't send notification to Slack channel!")



from datetime import datetime


class DebugLogCat:
    @staticmethod
    def log(debug, class_name, message):
        if debug:
            print("[%s - %s] -> %s" % (datetime.now().__str__(), class_name, message))



import matplotlib

matplotlib.use('Agg')

from typing import Any, AnyStr
from matplotlib.figure import Figure
from datetime import datetime
from matplotlib.pyplot import title
from experiment_automator.constants import ExperimentConstants, SlackConstants
from textwrap import wrap


class ResultContainer(dict):
    def __init__(self, attrs):
        super().__init__({
            ExperimentConstants.KEY_EXPERIMENT_RESULTS: {},
            SlackConstants.KEY_SLACK_PAYLOAD: {
                ExperimentConstants.KEY_EXPERIMENT_RESULTS: {},
            }
        })
        self.attrs = attrs

    def add_model_result(self, result_label: AnyStr, value: Any):
        self.get_experiment_results_payload().update({result_label: value})
        self.get_slack_payload().get(ExperimentConstants.KEY_EXPERIMENT_RESULTS).update({result_label: value})

    def __save_figure(self, image_label: AnyStr, figure: Figure):
        path = "%s-%s.png" % (image_label, datetime.now().strftime("%H:%M:%S"))
        title("\n".join(wrap("%s" % list(self.attrs.values()))))
        figure.savefig(path)
        figure.clear()

        return path

    def __add_slack_attachment_image(self, image_label, image_path):
        if self.get_slack_payload().get(SlackConstants.KEY_SLACK_IMAGE_ATTACHMENTS, None) is None:
            self.get_slack_payload()[SlackConstants.KEY_SLACK_IMAGE_ATTACHMENTS] = []

        self.get_slack_payload()[SlackConstants.KEY_SLACK_IMAGE_ATTACHMENTS].append((image_label, image_path))

    def __add_to_storage(self, path: AnyStr):
        # TODO Drive result will be added when Drive integration completed
        pass

    def add_figure(self, image_label: AnyStr, figure: Figure):
        path = self.__save_figure(image_label, figure)

        # Add figure to Slack as additional attachment
        self.__add_slack_attachment_image(image_label, path)

        # Add figure to storage
        # TODO Adding figure to storage

    def set_slack_main_figure(self, image_label: AnyStr, figure: Figure):
        self.get_slack_payload()[SlackConstants.KEY_SLACK_MAIN_IMAGE] = self.__save_figure(image_label, figure)

    def get_experiment_results_payload(self):
        return self.get(ExperimentConstants.KEY_EXPERIMENT_RESULTS)

    def get_slack_payload(self):
        return self.get(SlackConstants.KEY_SLACK_PAYLOAD)

    def __repr__(self):
        return super().__repr__()



import yaml


class Config:
    def __init__(self, config_filename):
        self.config = yaml.load(open(config_filename))

    def get_int(self, key):
        val = self.get(key)

        if not (val is None):
            return int(val)

        return val

    def get_float(self, key):
        val = self.get(key)

        if not (val is None):
            return float(val)

        return val

    def get_str(self, key):
        val = self.get(key)

        if not (val is None):
            return str(val)

        return val

    def get_dict(self, key):
        val = self.get(key)

        if not (val is None):
            return dict(val)

        return val

    def get(self, key):
        return self.config.get(key, None)



import setuptools
import yaml

config = yaml.load(open("config.yml"))

with open(config["long_description_path"], "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=config["name"],
    version=config["version"],
    author=config["author"],
    author_email=config["author_email"],
    description=config["description"],
    long_description=long_description,
    long_description_content_type=config["long_description_content_type"],
    url=config["url"],
    packages=setuptools.find_packages(),
    classifiers=config["classifiers"]
)

# TODO install_requires will be added


class ConfigNotFoundException(Exception):
    def __init__(self, message):
        super().__init__(message)


class WorkingDirectoryDoesNotExistException(Exception):
    def __init__(self, message):
        super().__init__(message)


class ImageUploaderException(Exception):
    def __init__(self, message):
        super().__init__(message)


class OAuthException(Exception):
    def __init__(self, message):
        super().__init__(message)


from experiment_automator.constants import ExperimentConstants, CSVReporterConstants
from experiment_automator.exceptions import *
from experiment_automator.utils import DebugLogCat
from os.path import exists
from os import makedirs
from datetime import datetime


class CSVReporter:
    def __init__(self, debug, experiment_config, csv_config):
        self.debug = debug
        self.report_dir = None
        self.report_filename = None
        self.column_names = None
        self.header_written = False
        self.experiment_config = experiment_config
        self.csv_config = csv_config

    def __class_name(self):
        return self.__class__.__name__

    def __initialize_csv_report(self):
        DebugLogCat.log(self.debug, self.__class_name(), "Creating working directory for reporting results!")

        work_dir = self.experiment_config.get(ExperimentConstants.KEY_EXPERIMENT_WORKDIR, None)
        experiment_name = self.experiment_config.get(ExperimentConstants.KEY_EXPERIMENT_NAME, None)

        if experiment_name is None:
            raise ConfigNotFoundException("%s - Experiment name not found in config!" % self.__class_name())

        if not exists(work_dir):
            raise WorkingDirectoryDoesNotExistException("%s - Working directory \"%s\" does not exist! Set appropriate working directory!" % (self.__class_name(), work_dir))

        report_dir_path = experiment_name + "/"

        if not (work_dir is None):
            report_dir_path = work_dir + experiment_name + "/"

        if not exists(report_dir_path):
            makedirs(report_dir_path)

        self.report_dir = report_dir_path

        DebugLogCat.log(self.debug, self.__class_name(), "Working directory \"%s\" is created!" % self.report_dir)

        now = datetime.now()
        self.report_filename = "%s%s-%s_%s.csv" % (self.report_dir, experiment_name, now.date(), now.time().strftime("%H:%M:%S"))

        DebugLogCat.log(self.debug, self.__class_name(), "Report file for experiment \"%s\" is created!" % self.report_filename)

    def __write_values_with_separator(self, report_file, values, separator, fn):
        index = 0

        for column_name in self.column_names:
            fn(report_file, column_name, values)

            if index < len(self.column_names) - 1:
                report_file.write(separator)

            index += 1

        report_file.write("\n")

    def __write_to_csv(self, message, columns=None, separator=","):
        with open(self.report_filename, "a+") as report_file:
            if columns is not None:
                message = {key: message[key] for key in message.keys() if key in columns}

            DebugLogCat.log(self.debug, self.__class_name(), "Writing message \"%s\" to file \"%s\"" % (str(message), self.report_filename))

            # If header is not written, write header
            if not self.header_written:
                self.column_names = message.keys()
                self.__write_values_with_separator(report_file,
                                                   None,
                                                   separator,
                                                   fn=lambda file, column_name, values: report_file.write(str(column_name)))
                self.header_written = True

            self.__write_values_with_separator(report_file,
                                               message,
                                               separator,
                                               fn=lambda file, column_name, values: report_file.write(str(values.get(column_name,"-"))))

    def save_results_to_csv(self, results):
        csv_log_status = self.csv_config.get(CSVReporterConstants.KEY_CSV_STATUS, None)

        if not (self.csv_config is None) and (csv_log_status == CSVReporterConstants.VALUE_CSV_ENABLED):
            DebugLogCat.log(self.debug, self.__class_name(), "CSV is configured properly. Logging results...!")

            if self.report_dir is None:
                self.__initialize_csv_report()

            csv_format_config = self.csv_config.get(CSVReporterConstants.KEY_CSV_FORMAT, None)

            if not (csv_format_config is None):
                DebugLogCat.log(self.debug, self.__class_name(), "Founded csv format defined by user!")

                self.__write_to_csv(results,
                                    columns=csv_format_config.get(CSVReporterConstants.KEY_CSV_FORMAT_COLUMNS),
                                    separator=csv_format_config.get(CSVReporterConstants.KEY_CSV_FORMAT_SEPARATOR))
            else:
                DebugLogCat.log(self.debug, self.__class_name(), "There is no csv format defined by user. Writing with default csv format!")
                self.__write_to_csv(results)
        else:
            DebugLogCat.log(self.debug, self.__class_name(), "CSV is not configured so CSV logging is disabled!")



class ExperimentConstants:
    KEY_EXPERIMENT = "experiment"
    KEY_EXPERIMENT_NAME = "name"
    KEY_EXPERIMENT_PARAMS = "params"
    KEY_EXPERIMENT_PARAM_RANGE = "range"
    KEY_EXPERIMENT_WORKDIR = "workdir"
    KEY_EXPERIMENT_RESULTS = "results"


class SlackConstants:
    KEY_SLACK = "slack"
    KEY_SLACK_NOTIFICATION_FORMAT = "notification_format"
    KEY_SLACK_WEBHOOK_URL = "webhook_url"
    KEY_SLACK_NOTIFICATION_SUCCESS = "success"
    KEY_SLACK_NOTIFICATION_FAIL = "fail"
    KEY_SLACK_NOTIFICATION_PRETEXT = "pretext"
    KEY_SLACK_NOTIFICATION_TITLE = "title"
    KEY_SLACK_NOTIFICATION_TITLE_LINK = "title_link"
    KEY_SLACK_NOTIFICATION_THUMB_URL = "thumb_url"
    KEY_SLACK_NOTIFICATION_IMAGE_URL = "image_url"
    KEY_SLACK_NOTIFICATION_COLOR = "color"
    KEY_SLACK_NOTIFICATION_FOOTER = "footer"
    KEY_SLACK_NOTIFICATION_TS = "ts"
    KEY_SLACK_ATTACHMENTS = "attachments"
    KEY_SLACK_IMAGE_ATTACHMENTS = "image_attachments"
    KEY_SLACK_MAIN_IMAGE = "main_image"
    KEY_SLACK_FIELDS = "fields"
    KEY_SLACK_FIELD_TITLE = "title"
    KEY_SLACK_FIELD_VALUE = "value"
    KEY_SLACK_FIELD_SHORT = "short"
    KEY_SLACK_IMAGE_SERVICE = "image_service"
    KEY_SLACK_PAYLOAD = "slack_payload"
    KEY_SLACK_IMAGE_SERVICE_UPLOAD_URL = "upload_url"
    VALUE_SLACK_NOTIFICATION_COLORS = {
        KEY_SLACK_NOTIFICATION_SUCCESS: "#7CD197",
        KEY_SLACK_NOTIFICATION_FAIL: "#F35A00"
    }
    VALUE_SLACK_NOTIFICATION_PRETEXTS = {
        KEY_SLACK_NOTIFICATION_SUCCESS: "Model training successfully completed!",
        KEY_SLACK_NOTIFICATION_FAIL: "Model training failed!"
    }
    VALUE_SLACK_NOTIFICATION_TITLE_RESULTS = "================= RESULTS =================\n"
    VALUE_SLACK_NOTIFICATION_TITLE_PARAMS = "=============== PARAMETERS ===============\n"
    VALUE_SLACK_NOTIFICATION_FOOTER = "Experiment Automator"


class CSVReporterConstants:
    KEY_CSV = "csv"
    KEY_CSV_FILENAME = "filename"
    KEY_CSV_FORMAT = "csv_format"
    KEY_CSV_FORMAT_SEPARATOR = "separator"
    KEY_CSV_FORMAT_COLUMNS = "columns"
    KEY_CSV_STATUS = "status"
    VALUE_CSV_ENABLED = "enabled"
    VALUE_CSV_DISABLED = "disabled"


class OAuthConstants:
    KEY_OAUTH = "oauth"
    KEY_OAUTH_SERVICE_NAME = "service_name"
    KEY_OAUTH_CLIENT_KEY = "client_key"
    KEY_OAUTH_CLIENT_SECRET = "client_secret"
    KEY_OAUTH_REQUEST_TOKEN_URL = "request_token_url"
    KEY_OAUTH_ACCESS_TOKEN_URL = "access_token_url"
    KEY_OAUTH_AUTHORIZATION_URL = "authorization_url"
    KEY_OAUTH_BASE_URL = "base_url"
    KEY_OAUTH_CALLBACK = "oauth_callback"
    KEY_OAUTH_VERIFIER = "oauth_verifier"
    KEY_OAUTH_TOKEN = "oauth_token"
    KEY_OAUTH_TOKEN_SECRET = "oauth_token_secret"
    KEY_OAUTH_ACCESS_TOKEN_PATH = "access_token_path"
    KEY_OAUTH_CALLBACK_OOB = "oob"


class DriveConstants:
    KEY_DRIVE = "drive"


class FlickrConstants:
    FLICKR_OAUTH_REQUEST_TOKEN_URL = "https://www.flickr.com/services/oauth/request_token"
    FLICKR_OAUTH_AUTHORIZATION_URL = "https://www.flickr.com/services/oauth/authorize"
    FLICKR_OAUTH_ACCESS_TOKEN_URL = "https://www.flickr.com/services/oauth/access_token"
    FLICKR_OAUTH_BASE_URL = "https://api.flickr.com/services/rest"
    FLICKR_UPLOAD_URL = "https://up.flickr.com/services/upload/"
    FLICKR_GET_IMAGE_SIZES_METHOD = "flickr.photos.getSizes"
    FLICKR_IMAGE_TYPE_SQUARE = "Square"
    FLICKR_IMAGE_TYPE_LARGE_SQUARE = "Large Square"
    FLICKR_IMAGE_TYPE_THUMBNAIL = "Thumbnail"
    FLICKR_IMAGE_TYPE_SMALL = "Small"
    FLICKR_IMAGE_TYPE_ORIGINAL = "Original"


class OtherConstants:
    KEY_DEBUG = "debug"
    KEY_DEBUG_TRACEBACK = "debug_traceback"
    KEY_START_TIME = "start_time"
    KEY_ELAPSED_TIME = "elapsed_time"
    KEY_COMPLETION_TIME = "completion_time"
    KEY_EXEC_STATUS = "exec_status"
    KEY_ERROR_CAUSE = "error_cause"
    EXECUTION_STATUS_SUCCESS = "SUCCESS"
    EXECUTION_STATUS_FAILED = "FAILED"
    IMAGE_SERVICE_FLICKR = "flickr"



from requests_oauthlib import OAuth1Session
from experiment_automator.constants import OAuthConstants
from experiment_automator.exceptions import OAuthException
from experiment_automator.utils import DebugLogCat
from os.path import exists


class OAuthClient:
    def __init__(self, debug, oauth_config):
        self.debug = debug
        self.oauth_config = oauth_config

        self.__check_config(OAuthConstants.KEY_OAUTH_SERVICE_NAME)
        self.__check_config(OAuthConstants.KEY_OAUTH_CLIENT_KEY)
        self.__check_config(OAuthConstants.KEY_OAUTH_CLIENT_SECRET)
        self.__check_config(OAuthConstants.KEY_OAUTH_REQUEST_TOKEN_URL)
        self.__check_config(OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_URL)
        self.__check_config(OAuthConstants.KEY_OAUTH_AUTHORIZATION_URL)
        self.__check_config(OAuthConstants.KEY_OAUTH_BASE_URL)
        self.__check_config(OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_PATH)

    def __class_name(self):
        return self.__class__.__name__

    def __check_config(self, config):
        if self.oauth_config.get(config, None) is None:
            raise OAuthException("%s - OAuth config %s is None. Please set in config!" % (self.__class_name(), config))

    def connect(self, callback, **params):
        global resource_owner_key
        global resource_owner_secret

        DebugLogCat.log(self.debug, self.__class_name(), "Getting authentication...")

        # If persist read oauth tokens from file
        if exists(self.oauth_config.get(OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_PATH)):
            DebugLogCat.log(self.debug, self.__class_name(), "User already authenticated. Read tokens from file...")

            with open(self.oauth_config.get(OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_PATH), "r") as key_file:
                resource_owner_key = key_file.readline().strip()
                resource_owner_secret = key_file.readline().strip()
        else:
            DebugLogCat.log(self.debug, self.__class_name(), "Getting request tokens from request token url...")

            oauth = OAuth1Session(client_key=self.oauth_config.get(OAuthConstants.KEY_OAUTH_CLIENT_KEY),
                                  client_secret=self.oauth_config.get(OAuthConstants.KEY_OAUTH_CLIENT_SECRET),
                                  callback_uri=callback)
            tokens = oauth.fetch_request_token(self.oauth_config.get(OAuthConstants.KEY_OAUTH_REQUEST_TOKEN_URL))
            resource_owner_key = tokens.get(OAuthConstants.KEY_OAUTH_TOKEN)
            resource_owner_secret = tokens.get(OAuthConstants.KEY_OAUTH_TOKEN_SECRET)

            DebugLogCat.log(self.debug, self.__class_name(), "User will be redirected to authorize the application...")

            authorization_url = oauth.authorization_url(self.oauth_config.get(OAuthConstants.KEY_OAUTH_AUTHORIZATION_URL), **params)
            print("Visit this URL in your browser for service \"{service_name}\": {url}".format(
                    service_name=self.oauth_config.get(OAuthConstants.KEY_OAUTH_SERVICE_NAME),
                    url=authorization_url))
            oauth_verifier = input("Enter PIN from browser: ")

            DebugLogCat.log(self.debug, self.__class_name(), "Getting access tokens from access token url...")

            oauth = OAuth1Session(client_key=self.oauth_config.get(OAuthConstants.KEY_OAUTH_CLIENT_KEY),
                                  client_secret=self.oauth_config.get(OAuthConstants.KEY_OAUTH_CLIENT_SECRET),
                                  resource_owner_key=resource_owner_key,
                                  resource_owner_secret=resource_owner_secret,
                                  verifier=oauth_verifier)

            oauth_tokens = oauth.fetch_access_token(self.oauth_config.get(OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_URL))
            resource_owner_key = oauth_tokens.get(OAuthConstants.KEY_OAUTH_TOKEN)
            resource_owner_secret = oauth_tokens.get(OAuthConstants.KEY_OAUTH_TOKEN_SECRET)

            DebugLogCat.log(self.debug, self.__class_name(), "Persist tokens in specified path...")

            # Persist OAuth key and secret
            with open(self.oauth_config.get(OAuthConstants.KEY_OAUTH_ACCESS_TOKEN_PATH), "w") as key_file:
                key_file.write(resource_owner_key + "\n")
                key_file.write(resource_owner_secret + "\n")

            DebugLogCat.log(self.debug, self.__class_name(), "Session established!")

        return OAuth1Session(client_key=self.oauth_config.get(OAuthConstants.KEY_OAUTH_CLIENT_KEY),
                             client_secret=self.oauth_config.get(OAuthConstants.KEY_OAUTH_CLIENT_SECRET),
                             resource_owner_key=resource_owner_key,
                             resource_owner_secret=resource_owner_secret)


